/*
 * Smart Ring Firmware for nRF52832
 * - Streams IMU (ICM-42670P) and PPG (MAX30102) data over BLE
 * - Uses Nordic SDK and SoftDevice
 * - I2C at 400kHz via TWI
 * - BLE GATT server with motion and heart rate characteristics
 * - Power management with System ON sleep and event-driven wakeup
*/

#include <stdbool.h>
#include <stdint.h>
#include "nrf.h"
#include "app_error.h"
#include "nrf_drv_twi.h"
#include "nrf_drv_clock.h"
#include "app_timer.h"
#include "ble.h"
#include "ble_srv_common.h"
#include "ble_advdata.h"
#include "ble_conn_params.h"
#include "nrf_sdh.h"
#include "nrf_sdh_ble.h"
#include "nrf_pwr_mgmt.h"
#include "nrf_delay.h"

// Defines
#define DEVICE_NAME                     "SmartRing"
#define MOTION_SERVICE_UUID_BASE        {0x23,0x15,0xA4,0xC1,0xD8,0x43,0x11,0xE8,0xB4,0x2B,0x23,0x66,0x00,0x00,0x00,0x00}
#define MOTION_SERVICE_UUID            0x1400
#define MOTION_CHAR_UUID_ACCEL         0x1401
#define MOTION_CHAR_UUID_GYRO          0x1402
#define HR_SERVICE_UUID_BASE            {0x23,0x15,0xA4,0xC1,0xD8,0x43,0x11,0xE8,0xB4,0x2B,0x23,0x66,0x10,0x00,0x00,0x00}
#define HR_SERVICE_UUID                0x2A37

#define TWI_INSTANCE_ID                 0
#define I2C_SCL_PIN                     27
#define I2C_SDA_PIN                     26

#define IMU_ADDR                        0x68
#define PPG_ADDR                        0x57

#define APP_BLE_OBSERVER_PRIO           3
#define APP_BLE_CONN_CFG_TAG            1
#define APP_ADV_INTERVAL                64  // 40 ms
#define APP_ADV_DURATION                18000  // 180 seconds

// Globals
static nrf_drv_twi_t m_twi = NRF_DRV_TWI_INSTANCE(TWI_INSTANCE_ID);
static uint16_t m_conn_handle = BLE_CONN_HANDLE_INVALID;
static ble_gatts_char_handles_t m_accel_handles;
static ble_gatts_char_handles_t m_gyro_handles;
static ble_gatts_char_handles_t m_hr_handles;

// Forward declarations
static void ble_stack_init(void);
static void gap_params_init(void);
static void gatt_init(void);
static void services_init(void);
static void advertising_init(void);
static void conn_params_init(void);
static void twi_init(void);
static void sensor_init(void);
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context);
static void idle_state_handle(void);

// BLE Event Handler
static void ble_evt_handler(ble_evt_t const * p_ble_evt, void * p_context)
{
    ret_code_t err_code;
    switch (p_ble_evt->header.evt_id)
    {
        case BLE_GAP_EVT_CONNECTED:
            m_conn_handle = p_ble_evt->evt.gap_evt.conn_handle;
            break;
        case BLE_GAP_EVT_DISCONNECTED:
            m_conn_handle = BLE_CONN_HANDLE_INVALID;
            break;
        default:
            break;
    }
}

// BLE Init
static void ble_stack_init(void)
{
    ret_code_t err_code;
    err_code = nrf_sdh_enable_request();
    APP_ERROR_CHECK(err_code);

    uint32_t ram_start = 0;
    err_code = nrf_sdh_ble_default_cfg_set(APP_BLE_CONN_CFG_TAG, &ram_start);
    APP_ERROR_CHECK(err_code);

    err_code = nrf_sdh_ble_enable(&ram_start);
    APP_ERROR_CHECK(err_code);

    NRF_SDH_BLE_OBSERVER(m_ble_observer, APP_BLE_OBSERVER_PRIO, ble_evt_handler, NULL);
}

static void gap_params_init(void)
{
    ret_code_t err_code;
    ble_gap_conn_sec_mode_t sec_mode;
    BLE_GAP_CONN_SEC_MODE_SET_OPEN(&sec_mode);

    err_code = sd_ble_gap_device_name_set(&sec_mode,
                                         (const uint8_t *)DEVICE_NAME,
                                         strlen(DEVICE_NAME));
    APP_ERROR_CHECK(err_code);

    ble_gap_conn_params_t gap_conn_params;
    memset(&gap_conn_params, 0, sizeof(gap_conn_params));

    gap_conn_params.min_conn_interval = MSEC_TO_UNITS(20, UNIT_1_25_MS);
    gap_conn_params.max_conn_interval = MSEC_TO_UNITS(75, UNIT_1_25_MS);
    gap_conn_params.slave_latency     = 0;
    gap_conn_params.conn_sup_timeout  = MSEC_TO_UNITS(4000, UNIT_10_MS);

    err_code = sd_ble_gap_ppcp_set(&gap_conn_params);
    APP_ERROR_CHECK(err_code);
}

static void services_init(void)
{
    ret_code_t err_code;
    ble_uuid_t service_uuid;
    ble_uuid128_t base_uuid = MOTION_SERVICE_UUID_BASE;

    // Add custom motion service
    err_code = sd_ble_uuid_vs_add(&base_uuid, &service_uuid.type);
    APP_ERROR_CHECK(err_code);
    service_uuid.uuid = MOTION_SERVICE_UUID;

    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                        &service_uuid,
                                        NULL);
    APP_ERROR_CHECK(err_code);

    // Add accel char
    {
        ble_gatts_char_md_t char_md = {0};
        char_md.char_props.notify = 1;

        ble_gatts_attr_md_t cccd_md = {0};
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);
        cccd_md.vloc = BLE_GATTS_VLOC_STACK;
        char_md.p_cccd_md = &cccd_md;

        ble_uuid_t char_uuid;
        char_uuid.type = service_uuid.type;
        char_uuid.uuid = MOTION_CHAR_UUID_ACCEL;

        ble_gatts_attr_md_t attr_md = {0};
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
        attr_md.vloc       = BLE_GATTS_VLOC_STACK;

        ble_gatts_attr_t    attr_char_value = {0};
        attr_char_value.p_uuid    = &char_uuid;
        attr_char_value.p_attr_md = &attr_md;
        attr_char_value.max_len   = sizeof(int16_t) * 3;
        attr_char_value.init_len  = sizeof(int16_t) * 3;

        err_code = sd_ble_gatts_characteristic_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                                   &char_md,
                                                   &attr_char_value,
                                                   &m_accel_handles);
        APP_ERROR_CHECK(err_code);
    }

    // Add gyro char (similar to accel)
    {
        ble_gatts_char_md_t char_md = {0};
        char_md.char_props.notify = 1;
        char_md.p_cccd_md        = &((ble_gatts_attr_md_t){.read_perm = {1,1}, .write_perm = {1,1}, .vloc=BLE_GATTS_VLOC_STACK});

        ble_uuid_t char_uuid;
        char_uuid.type = service_uuid.type;
        char_uuid.uuid = MOTION_CHAR_UUID_GYRO;

        ble_gatts_attr_md_t attr_md = {0};
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
        attr_md.vloc       = BLE_GATTS_VLOC_STACK;

        ble_gatts_attr_t attr_char_value = {0};
        attr_char_value.p_uuid    = &char_uuid;
        attr_char_value.p_attr_md = &attr_md;
        attr_char_value.max_len   = sizeof(int16_t) * 3;
        attr_char_value.init_len  = sizeof(int16_t) * 3;

        err_code = sd_ble_gatts_characteristic_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                                   &char_md,
                                                   &attr_char_value,
                                                   &m_gyro_handles);
        APP_ERROR_CHECK(err_code);
    }

    // Add standard Heart Rate service
    ble_uuid_t hr_uuid;
    ble_uuid128_t hr_base = HR_SERVICE_UUID_BASE;
    err_code = sd_ble_uuid_vs_add(&hr_base, &hr_uuid.type);
    APP_ERROR_CHECK(err_code);
    hr_uuid.uuid = HR_SERVICE_UUID;

    err_code = sd_ble_gatts_service_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                        &hr_uuid,
                                        NULL);
    APP_ERROR_CHECK(err_code);

    // Add HR measurement characteristic
    {
        ble_gatts_char_md_t char_md = {0};
        char_md.char_props.notify = 1;

        ble_gatts_attr_md_t cccd_md = {0};
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.read_perm);
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&cccd_md.write_perm);
        cccd_md.vloc = BLE_GATTS_VLOC_STACK;
        char_md.p_cccd_md = &cccd_md;

        ble_gatts_attr_md_t attr_md = {0};
        BLE_GAP_CONN_SEC_MODE_SET_OPEN(&attr_md.read_perm);
        attr_md.vloc       = BLE_GATTS_VLOC_STACK;

        ble_uuid_t char_uuid = hr_uuid;

        ble_gatts_attr_t    attr_char_value = {0};
        attr_char_value.p_uuid    = &char_uuid;
        attr_char_value.p_attr_md = &attr_md;
        attr_char_value.max_len   = sizeof(uint8_t);
        attr_char_value.init_len  = sizeof(uint8_t);

        err_code = sd_ble_gatts_characteristic_add(BLE_GATTS_SRVC_TYPE_PRIMARY,
                                                   &char_md,
                                                   &attr_char_value,
                                                   &m_hr_handles);
        APP_ERROR_CHECK(err_code);
    }
}

static void advertising_init(void)
{
    ret_code_t err_code;
    ble_advdata_t advdata;
    ble_uuid_t adv_uuids[] = {{MOTION_SERVICE_UUID, BLE_UUID_TYPE_VENDOR_BEGIN}};

    memset(&advdata, 0, sizeof(advdata));

    advdata.name_type               = BLE_ADVDATA_FULL_NAME;
    advdata.include_appearance      = false;
    advdata.flags                   = BLE_GAP_ADV_FLAGS_LE_ONLY_GENERAL_DISC_MODE;
    advdata.uuids_complete.uuid_cnt = sizeof(adv_uuids) / sizeof(adv_uuids[0]);
    advdata.uuids_complete.p_uuids  = adv_uuids;

    err_code = ble_advdata_set(&advdata, NULL);
    APP_ERROR_CHECK(err_code);
}

static void conn_params_init(void)
{
    ble_conn_params_init_t cp_init;
    memset(&cp_init, 0, sizeof(cp_init));

    cp_init.p_conn_params                  = NULL;
    cp_init.first_conn_params_update_delay = APP_TIMER_TICKS(5000);
    cp_init.next_conn_params_update_delay  = APP_TIMER_TICKS(30000);
    cp_init.max_conn_params_update_count   = 3;
    cp_init.start_on_notify_cccd_handle    = BLE_GATT_HANDLE_INVALID;
    cp_init.disconnect_on_fail             = true;
    cp_init.evt_handler                    = NULL;
    cp_init.error_handler                  = APP_ERROR_HANDLER;

    ret_code_t err_code = ble_conn_params_init(&cp_init);
    APP_ERROR_CHECK(err_code);
}

// TWI init
static void twi_init(void)
{
    ret_code_t err_code;
    const nrf_drv_twi_config_t twi_config = {
       .scl                = I2C_SCL_PIN,
       .sda                = I2C_SDA_PIN,
       .frequency          = NRF_DRV_TWI_FREQ_400K,
       .interrupt_priority = APP_IRQ_PRIORITY_HIGH,
       .clear_bus_init     = false
    };

    err_code = nrf_drv_twi_init(&m_twi, &twi_config, NULL, NULL);
    APP_ERROR_CHECK(err_code);

    nrf_drv_twi_enable(&m_twi);
}

// Sensor initialization
static void sensor_init(void)
{
    // IMU: set register configs (power, ranges)
    // Example: write to IMU_ADDR via TWI
    uint8_t imu_init_data[] = {0x6B, 0x00}; // Wake from sleep
    APP_ERROR_CHECK(nrf_drv_twi_tx(&m_twi, IMU_ADDR, imu_init_data, sizeof(imu_init_data), false));

    // PPG: reset and setup LED currents, modes
    uint8_t ppg_reset[] = {0x09, 0x40};
    APP_ERROR_CHECK(nrf_drv_twi_tx(&m_twi, PPG_ADDR, ppg_reset, sizeof(ppg_reset), false));
}

int main(void)
{
    // Initialize clocks
    nrf_drv_clock_init();
    nrf_drv_clock_lfclk_request(NULL);

    // Init modules
    twi_init();
    ble_stack_init();
    gap_params_init();
    services_init();
    advertising_init();
    conn_params_init();
    sensor_init();
    nrf_pwr_mgmt_init();

    // Start advertising
    sd_ble_gap_adv_start(BLE_GAP_ADV_SET_HANDLE_NOT_SET, APP_BLE_CONN_CFG_TAG);

    // Main loop
    while (true)
    {
        // Read sensors
        int16_t accel_data[3];
        int16_t gyro_data[3];
        uint8_t hr_data;
        
        // Read IMU via TWI registers (example registers)
        // nrf_drv_twi_rx(&m_twi, IMU_ADDR, (uint8_t*)accel_data, sizeof(accel_data));

        // Read PPG
        // nrf_drv_twi_rx(&m_twi, PPG_ADDR, &hr_data, sizeof(hr_data));

        // Notify BLE clients
        if (m_conn_handle != BLE_CONN_HANDLE_INVALID)
        {
            ble_gatts_hvx_params_t hvx;
            memset(&hvx, 0, sizeof(hvx));

            // Send accel
            hvx.handle = m_accel_handles.value_handle;
            hvx.p_data = (uint8_t*)accel_data;
            hvx.p_len  = (uint16_t[]){sizeof(accel_data)};
            hvx.type   = BLE_GATT_HVX_NOTIFICATION;
            sd_ble_gatts_hvx(m_conn_handle, &hvx);

            // Send gyro
            hvx.handle = m_gyro_handles.value_handle;
            hvx.p_data = (uint8_t*)gyro_data;
            hvx.p_len  = (uint16_t[]){sizeof(gyro_data)};
            sd_ble_gatts_hvx(m_conn_handle, &hvx);

            // Send HR
            hvx.handle = m_hr_handles.value_handle;
            hvx.p_data = &hr_data;
            hvx.p_len  = (uint16_t[]){sizeof(hr_data)};
            sd_ble_gatts_hvx(m_conn_handle, &hvx);
        }

        // Power management
        nrf_pwr_mgmt_run();
    }
}

/** @} */
